# -*- coding: utf-8 -*-
from __future__ import absolute_import  
import functools
from scrapy.utils.serialize import ScrapyJSONEncoder

from tornado import gen
from twisted.internet.defer import Deferred
from twisted.internet import reactor


def tt_coroutine(func):
    """
    Twisted-Tornado coroutine which uses tornago.gen inside,
    but returns Twisted Deferred to the outside.
    """
    return gen_to_twisted(gen.coroutine(func))


def wrap_future(future):
    """
    Wrap tornado.concurrent.Future in a twisted.internet.defer.Deferred.
    Only basics are supported: callback, errback.
    """
    d = Deferred()

    def callback(future):
        e = future.exception()
        if e:
            reactor.callFromThread(d.errback, e)
            return
        reactor.callFromThread(d.callback, future.result())

    future.add_done_callback(callback)
    return d


def gen_to_twisted(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return wrap_future(func(*args, **kwargs))
    return wrapper


# XXX: this is copy-pasted to make motor_exporter independent
_encoder = ScrapyJSONEncoder(ensure_ascii=False)
def json_encode(obj):
    """
    Encode a Python object to JSON.
    Unlike standard json.dumps, datetime.datetime objects are supported.
    """
    return _encoder.encode(obj)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import logging
import itertools
import operator
from scrapy.core.downloader import Downloader
from scrapy.utils.reactor import CallLaterOnce
import six

from scrapy import signals
from scrapy.signalmanager import SignalManager
from scrapy.crawler import CrawlerProcess, Crawler
from scrapy.core.engine import ExecutionEngine
# from scrapy.utils.engine import get_engine_status

from arachnado.signals import Signal
from arachnado import stats
from arachnado.process_stats import ProcessStatsMonitor

logger = logging.getLogger(__name__)

# monkey patch Scrapy to add an extra signal
signals.spider_closing = object()
signals.engine_paused = object()
signals.engine_resumed = object()
signals.engine_tick = object()
signals.downloader_enqueued = object()
signals.downloader_dequeued = object()


# a signal which is fired when stats are changed in any of the spiders
agg_stats_changed = Signal("agg_stats_changed", False)
STAT_SIGNALS = {
    stats.stats_changed: agg_stats_changed,
}


SCRAPY_SIGNAL_NAMES = [
    'engine_started',
    'engine_stopped',
    'engine_paused',  # custom
    'engine_resumed',  # custom
    'engine_tick',  # custom
    'item_scraped',
    'item_dropped',
    'spider_closed',
    'spider_closing',  # custom
    'spider_opened',
    'spider_idle',
    'spider_error',
    'request_scheduled',
    'request_dropped',
    'response_received',
    'response_downloaded',
    'downloader_enqueued',  # custom
    'downloader_dequeued',  # custom
]


def _get_crawler_process_signals_cls():
    spider_to_cp = {}

    class CrawlerProcessSignals(object):
        @classmethod
        def signal(cls, spider_signal):
            return spider_to_cp[spider_signal]

        engine_started = Signal('engine_started', True)
        engine_stopped = Signal('engine_stopped', True)
        engine_paused = Signal('engine_paused', False)  # custom
        engine_resumed = Signal('engine_resumed', False)  # custom
        engine_tick = Signal('engine_tick', False)  # custom
        spider_opened = Signal('spider_opened', True)
        spider_idle = Signal('spider_idle', False)
        spider_closed = Signal('spider_closed', True)
        spider_closing = Signal('spider_closing', False)  # custom
        spider_error = Signal('spider_error', False)
        request_scheduled = Signal('request_scheduled', False)
        request_dropped = Signal('request_dropped', False)
        response_received = Signal('response_received', False)
        response_downloaded = Signal('response_downloaded', False)
        item_scraped = Signal('item_scraped', True)
        item_dropped = Signal('item_dropped', True)
        downloader_enqueued = Signal('downloader_enqueued', False)
        downloader_dequeued = Signal('downloader_dequeued', False)

    for name in SCRAPY_SIGNAL_NAMES:
        signal = getattr(signals, name)
        cp_signal = getattr(CrawlerProcessSignals, name)
        spider_to_cp[signal] = cp_signal

    return CrawlerProcessSignals


CrawlerProcessSignals = _get_crawler_process_signals_cls()


class ArachnadoExecutionEngine(ExecutionEngine):
    """
    Extended ExecutionEngine.
    It sends a signal when engine gets scheduled to stop.
    """
    def __init__(self, *args, **kwargs):
        super(ArachnadoExecutionEngine, self).__init__(*args, **kwargs)
        self.send_tick = CallLaterOnce(self._send_tick_signal)

    def close_spider(self, spider, reason='cancelled'):
        if self.slot.closing:
            return self.slot.closing
        self.crawler.crawling = False
        self.signals.send_catch_log(signals.spider_closing)
        return super(ArachnadoExecutionEngine, self).close_spider(spider, reason)

    def pause(self):
        """Pause the execution engine"""
        super(ArachnadoExecutionEngine, self).pause()
        self.signals.send_catch_log(signals.engine_paused)

    def unpause(self):
        """Resume the execution engine"""
        super(ArachnadoExecutionEngine, self).unpause()
        self.signals.send_catch_log(signals.engine_resumed)

    def _next_request(self, spider):
        res = super(ArachnadoExecutionEngine, self)._next_request(spider)
        self.send_tick.schedule(0.1)  # avoid sending the signal too often
        return res

    def _send_tick_signal(self):
        self.signals.send_catch_log_deferred(signals.engine_tick)


class ArachnadoCrawler(Crawler):
    """
    Extended Crawler which uses ArachnadoExecutionEngine.
    """
    def _create_engine(self):
        return ArachnadoExecutionEngine(self, lambda _: self.stop())


class ArachnadoDownloader(Downloader):
    def _enqueue_request(self, request, spider):
        dfd = super(ArachnadoDownloader, self)._enqueue_request(request, spider)
        self.signals.send_catch_log(signals.downloader_enqueued)

        def _send_dequeued(_):
            self.signals.send_catch_log(signals.downloader_dequeued)
            return _

        dfd.addBoth(_send_dequeued)
        return dfd


class ArachnadoCrawlerProcess(CrawlerProcess):
    """
    CrawlerProcess which sets up a global signals manager,
    assigns unique ids to each spider job, workarounds some Scrapy
    issues and provides extra stats.
    """
    crawl_ids = itertools.count(start=1)

    def __init__(self, settings=None):
        self.signals = SignalManager(self)
        self.signals.connect(self.on_spider_closed, CrawlerProcessSignals.spider_closed)
        self._finished_jobs = []
        self._paused_jobs = set()
        self.procmon = ProcessStatsMonitor()
        self.procmon.start()

        super(ArachnadoCrawlerProcess, self).__init__(settings or {})

        # don't log DepthMiddleware messages
        # see https://github.com/scrapy/scrapy/issues/1308
        logging.getLogger("scrapy.spidermiddlewares.depth").setLevel(logging.INFO)

    def crawl(self, crawler_or_spidercls, *args, **kwargs):
        kwargs['crawl_id'] = next(self.crawl_ids)

        crawler = crawler_or_spidercls
        if not isinstance(crawler_or_spidercls, Crawler):
            crawler = self._create_crawler(crawler_or_spidercls)

        # aggregate all crawler signals
        for name in SCRAPY_SIGNAL_NAMES:
            crawler.signals.connect(self._resend_signal, getattr(signals, name))

        # aggregate signals from crawler EventedStatsCollectors
        if hasattr(crawler.stats, "signals"):
            crawler.stats.signals.connect(self._resend_signal, stats.stats_changed)

        d = super(ArachnadoCrawlerProcess, self).crawl(crawler_or_spidercls, *args, **kwargs)
        return d

    def _create_crawler(self, spidercls):
        if isinstance(spidercls, six.string_types):
            spidercls = self.spider_loader.load(spidercls)
        return ArachnadoCrawler(spidercls, self.settings)

    def stop_job(self, crawl_id):
        """ Stop a single crawl job """
        self.get_crawler(crawl_id).stop()

    def pause_job(self, crawl_id):
        """ Pause a crawling job """
        self._paused_jobs.add(crawl_id)
        self.get_crawler(crawl_id).engine.pause()

    def resume_job(self, crawl_id):
        """ Resume a crawling job """
        self._paused_jobs.remove(crawl_id)
        self.get_crawler(crawl_id).engine.unpause()

    def get_crawler(self, crawl_id):
        for crawler in self.crawlers:
            if getattr(crawler.spider, "crawl_id") == crawl_id:
                return crawler
        raise KeyError("Job is not known: %s" % crawl_id)

    def _resend_signal(self, **kwargs):
        # FIXME: this is a mess. Signal handling should be unified somehow:
        # there shouldn't be two separate code paths
        # for CrawlerProcessSignals and STAT_SIGNALS.
        signal = kwargs['signal']
        if signal in STAT_SIGNALS:
            signal = STAT_SIGNALS[signal]
            kwargs['crawler'] = kwargs.pop('sender').crawler
        else:
            signal = CrawlerProcessSignals.signal(signal)
            kwargs['crawler'] = kwargs.pop('sender')

        kwargs['signal'] = signal
        if signal.supports_defer:
            return self.signals.send_catch_log_deferred(**kwargs)
        else:
            return self.signals.send_catch_log(**kwargs)

    def stop(self):
        """ Terminate the process (exit from application). """
        self.procmon.stop()
        return super(ArachnadoCrawlerProcess, self).stop()

    def on_spider_closed(self, spider, reason):
        # spiders are closed not that often, insert(0,...) should be fine
        self._finished_jobs.insert(0, {
            'id': spider.crawl_id,
            'job_id': getattr(spider, 'motor_job_id'),
            'seed': spider.domain,
            'status': reason,
            'stats': spider.crawler.stats.get_stats(spider),
            'downloads': self._downloader_stats(spider.crawler)
        })

    # FIXME: methods below are ugly for two reasons:
    # 1. they assume spiders have certain attributes;
    # 2. they try to get crawling status based on auxilary information.

    def get_jobs(self):
        """ Return a list of active jobs """
        crawlers = [crawler for crawler in self.crawlers
                    if crawler.spider is not None]
        return [
            {
                'id': crawler.spider.crawl_id,
                'job_id': getattr(crawler.spider, 'motor_job_id'),
                'seed': crawler.spider.domain,
                'status': self._get_crawler_status(crawler),
                'stats': crawler.spider.crawler.stats.get_stats(crawler.spider),
                'downloads': self._downloader_stats(crawler)
                # 'engine_info': dict(get_engine_status(crawler.engine))
            }
            for crawler in crawlers
        ]

    @classmethod
    def _downloader_stats(cls, crawler):
        downloader = crawler.engine.downloader
        return {
            'active': [cls._request_info(req) for req in downloader.active],
            'slots': sorted([
                cls._slot_info(key, slot)
                for key, slot in downloader.slots.items()
            ], key=operator.itemgetter('key'))
        }

    @classmethod
    def _request_info(cls, request):
        return {'url': request.url, 'method': request.method}

    @classmethod
    def _slot_info(cls, key, slot):
        return {
            'key': key,
            'concurrency': slot.concurrency,
            'delay': slot.delay,
            'lastseen': slot.lastseen,
            'len(queue)': len(slot.queue),
            'transferring': [cls._request_info(req) for req in slot.transferring],
            'active': [cls._request_info(req) for req in slot.active],
        }

    def _get_crawler_status(self, crawler):
        if crawler.spider is None:
            return "unknown"
        if not crawler.crawling:
            return "stopping"
        if int(crawler.spider.crawl_id) in self._paused_jobs:
            return "suspended"
        return "crawling"

    @property
    def jobs(self):
        """ Current crawl state """
        # filter out active jobs which are in fact finished
        finished_ids = {job['id'] for job in self._finished_jobs}
        active_jobs = [job for job in self.get_jobs()
                       if job['id'] not in finished_ids]

        return active_jobs + self._finished_jobs



# -*- coding: utf-8 -*-
""" Process monitor: CPU, RAM """
from __future__ import absolute_import
import os
import time
import logging

# import yappi
import psutil
from tornado.ioloop import PeriodicCallback
from scrapy.signalmanager import SignalManager

logger = logging.getLogger(__name__)


class ProcessStatsMonitor(object):
    """ A class which emits process stats periodically """

    signal_updated = object()

    def __init__(self, interval=1.0):
        self.signals = SignalManager(self)
        self.process = psutil.Process(os.getpid())
        self.interval = interval
        self._task = PeriodicCallback(self._emit, self.interval*1000)
        self._recent = {}

    def start(self):
        # yappi.start()
        self._task.start()

    def stop(self):
        self._task.stop()
        # stats = yappi.get_func_stats()
        # stats.sort('tsub', 'desc')
        # with open("func-stats.txt", 'wt') as f:
        #     stats.print_all(f, columns={
        #         0: ("name", 80),
        #         1: ("ncall", 10),
        #         2: ("tsub", 8),
        #         3: ("ttot", 8),
        #         4: ("tavg",8)
        #     })
        #
        # pstats = yappi.convert2pstats(stats)
        # pstats.dump_stats("func-stats.prof")

    def get_recent(self):
        return self._recent

    def _emit(self):
        cpu_times = self.process.cpu_times()
        ram_usage = self.process.memory_info()
        stats = {
            'ram_percent': self.process.memory_percent(),
            'ram_rss': ram_usage.rss,
            'ram_vms': ram_usage.vms,
            'cpu_percent': self.process.cpu_percent(),
            'cpu_time_user': cpu_times.user,
            'cpu_time_system': cpu_times.system,
            'num_fds': self.process.num_fds(),
            'context_switches': self.process.num_ctx_switches(),
            'num_threads': self.process.num_threads(),
            'server_time': int(time.time()*1000),
        }
        self._recent = stats
        self.signals.send_catch_log(self.signal_updated, stats=stats)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
from ConfigParser import SafeConfigParser

_ROOT = os.path.abspath(os.path.dirname(__file__))

FILENAMES = [
    os.path.join(_ROOT, 'settings', 'defaults.conf'),
    '/etc/arachnado.conf',
    os.path.expanduser('~/.config/arachnado.conf'),
    os.path.expanduser('~/.arachnado.conf'),
]


def load_settings(config_files=(), overrides=()):
    cp = SafeConfigParser()
    cp.read(FILENAMES + config_files)

    for section, option, value in overrides:
        if value is not None:
            if isinstance(value, bool):
                value = int(value)
            cp.set(section, option, str(value))

    return {
        section: dict(cp.items(section))
        for section in cp.sections()
    }


def ensure_bool(opts, section, name):
    opts[section][name] = bool(int(opts[section][name]))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import datetime
import logging

import scrapy
from scrapy.linkextractors import LinkExtractor
from scrapy.http.response.html import HtmlResponse

from .utils import MB, add_scheme_if_missing, get_netloc
from .crawler_process import ArachnadoCrawler


DEFAULT_SETTINGS = {
    'DEPTH_STATS_VERBOSE': True,
    'DEPTH_PRIORITY': -1,

    'BOT_NAME': 'arachnado',
    'USER_AGENT': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) '
                  'AppleWebKit/537.36 (KHTML, like Gecko) '
                  'Chrome/39.0.2148.0 Safari/537.36',

    'MEMUSAGE_ENABLED': True,
    'DOWNLOAD_MAXSIZE': 1 * MB,
    # see https://github.com/scrapy/scrapy/issues/1303
    # 'DOWNLOAD_WARNSIZE': 1 * MB,

    # 'CLOSESPIDER_PAGECOUNT': 30,  # for debugging
    'LOG_LEVEL': 'DEBUG',
    'TELNETCONSOLE_ENABLED': False,
    # 'CONCURRENT_REQUESTS': 100,

    'EXTENSIONS': {
        'scrapy.extensions.throttle.AutoThrottle': None,
        'arachnado.extensions.throttle.AutoThrottle': 0,
    },

    'AUTOTHROTTLE_ENABLED': True,
    'AUTOTHROTTLE_DEBUG': False,
    'AUTOTHROTTLE_START_DELAY': 5,
    'AUTOTHROTTLE_MAX_DELAY': 60,
    'AUTOTHROTTLE_TARGET_CONCURRENCY': 1.0,
    'CONCURRENT_REQUESTS_PER_DOMAIN': 4,
    'DOWNLOAD_DELAY': 0.3,  # min download delay

    'STATS_CLASS': 'arachnado.stats.EventedStatsCollector',
    'DOWNLOADER': 'arachnado.crawler_process.ArachnadoDownloader',

    # see https://github.com/scrapy/scrapy/issues/1054
    'DOWNLOAD_HANDLERS': {'s3': None},

    'ITEM_PIPELINES': {
        'arachnado.motor_exporter.pipelines.MotorPipeline': 100,
    },

    'MOTOR_PIPELINE_JOBID_KEY': '_job_id',

    'HTTPCACHE_ENABLED': False,
    # This storage is read-only. Responses are stored by MotorPipeline
    'HTTPCACHE_STORAGE': 'arachnado.extensions.httpcache.'
                         'ArachnadoCacheStorage',
}


def create_crawler(settings=None, spider_cls=None):
    _settings = DEFAULT_SETTINGS.copy()
    _settings.update(settings or {})
    return ArachnadoCrawler(spider_cls, _settings)


class ArachnadoSpider(scrapy.Spider):
    """
    A base spider that contains common attributes and utilities for all
    Arachnado spiders
    """
    crawl_id = None
    domain = None
    motor_job_id = None

    def __init__(self, *args, **kwargs):
        super(ArachnadoSpider, self).__init__(*args, **kwargs)
        # don't log scraped items
        logging.getLogger("scrapy.core.scraper").setLevel(logging.INFO)

    def get_page_item(self, response, type_='page'):
        return {
            'crawled_at': datetime.datetime.utcnow(),
            'url': response.url,
            'status': response.status,
            'headers': response.headers,
            'body': response.body_as_unicode(),
            'meta': response.meta,
            '_type': type_,
        }


class CrawlWebsiteSpider(ArachnadoSpider):
    """
    A spider which crawls all the website.
    To run it, set its ``crawl_id`` and ``domain`` arguments.
    """
    name = 'crawlwebsite'
    custom_settings = {
        'DEPTH_LIMIT': 10,
    }

    def __init__(self, *args, **kwargs):
        super(CrawlWebsiteSpider, self).__init__(*args, **kwargs)
        self.start_url = add_scheme_if_missing(self.domain)

    def start_requests(self):
        self.logger.info("Started job %s#%d for domain %s",
                         self.motor_job_id, self.crawl_id, self.domain)
        yield scrapy.Request(self.start_url, self.parse_first,
                             dont_filter=True)

    def parse_first(self, response):
        # If there is a redirect in the first request, use the target domain
        # to restrict crawl instead of the original.
        self.domain = get_netloc(response.url)
        self.crawler.stats.set_value('arachnado/start_url', self.start_url)
        self.crawler.stats.set_value('arachnado/domain', self.domain)

        allow_domain = self.domain
        if self.domain.startswith("www."):
            allow_domain = allow_domain[len("www."):]

        self.get_links = LinkExtractor(
            allow_domains=[allow_domain]
        ).extract_links

        for elem in self.parse(response):
            yield elem

    def parse(self, response):
        if not isinstance(response, HtmlResponse):
            self.logger.info("non-HTML response is skipped: %s" % response.url)
            return

        yield self.get_page_item(response)

        for link in self.get_links(response):
            yield scrapy.Request(link.url, self.parse)



# -*- coding: utf-8 -*-
"""
An interactive Python interpreter available through telnet.
"""
from __future__ import absolute_import
from twisted.conch.manhole import ColoredManhole
from twisted.conch.insults import insults
from twisted.conch.telnet import TelnetTransport, TelnetBootstrapProtocol
from twisted.internet import protocol


def start(port=None, host=None, telnet_vars=None):
    from twisted.internet import reactor

    port = int(port) if port else 6023
    host = host or '127.0.0.1'
    telnet_vars = telnet_vars or {}

    factory = protocol.ServerFactory()
    factory.protocol = lambda: TelnetTransport(
        TelnetBootstrapProtocol,
        insults.ServerProtocol,
        ColoredManhole,
        telnet_vars
    )
    return reactor.listenTCP(port, factory, interface=host)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import functools

from tornado.ioloop import PeriodicCallback
from scrapy.statscollectors import StatsCollector
from scrapy.signalmanager import SignalManager

from arachnado.signals import Signal
from arachnado.utils import decorate_methods


def store_changed_value(meth):
    @functools.wraps(meth)
    def wrapper(self, key, *args, **kwargs):
        old_value = self._stats.get(key)
        meth(self, key, *args, **kwargs)
        value = self._stats.get(key)
        if value != old_value:
            self._changes[key] = value
    return wrapper


def store_changed_stats(meth):
    @functools.wraps(meth)
    def wrapper(self, *args, **kwargs):
        meth(self, *args, **kwargs)
        self._changes = self._stats
    return wrapper


stats_changed = Signal("stats_changed", False)


@decorate_methods(["set_value", "inc_value", "max_value", "min_value"], store_changed_value)
@decorate_methods(["set_stats", "clear_stats"], store_changed_stats)
class EventedStatsCollector(StatsCollector):
    """
    Stats Collector which allows to subscribe to value changes.
    Update notifications are throttled: interval between updates is no shorter
    than ``accumulate_time``.

    It is assumed that stat keys are never deleted.
    """
    accumulate_time = 0.1  # value is in seconds

    def __init__(self, crawler):
        super(EventedStatsCollector, self).__init__(crawler)
        self.signals = SignalManager(self)
        self._changes = {}
        self._task = PeriodicCallback(self.emit_changes, self.accumulate_time*1000)
        self._task.start()

        # FIXME: this is ugly
        self.crawler = crawler  # used by ArachnadoCrawlerProcess

    def emit_changes(self):
        if self._changes:
            changes, self._changes = self._changes, {}
            self.signals.send_catch_log(stats_changed, changes=changes)

    def open_spider(self, spider):
        super(EventedStatsCollector, self).open_spider(spider)
        self._task.start()

    def close_spider(self, spider, reason):
        super(EventedStatsCollector, self).close_spider(spider, reason)
        self._task.stop()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import json
import logging

from tornado import websocket

from arachnado.utils import json_encode


logger = logging.getLogger(__name__)


class BaseWSHandler(websocket.WebSocketHandler):
    """
    A base class which knows how to communicate with a client.
    All messages are JSON-encoded objects ``{"event": name, "data": data}``.

    Subclasses should use ``on_event`` instead of ``on_message`` and
    ``write_event`` instead of ``write_message``.
    """

    def write_event(self, event, data):
        """ Send a message to the client """
        message = json_encode({'event': event, 'data': data})
        self.write_message(message)

    def on_message(self, message):
        try:
            msg = json.loads(message)
            event, data = msg['event'], msg['data']
        except Exception as e:
            logger.warn("Invalid message skipped" + message[:500])
            return
        self.on_event(event, data)

    def on_event(self, event, data):
        """ This method is called when a message is received from a client """
        pass

    def on_open(self, *args, **kwargs):
        """ ``open`` alias, for consistency """

    def open(self, *args, **kwargs):
        self.on_open(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
Async MongoDB item exporter using Motor_.

.. _Motor: https://github.com/mongodb/motor
"""
from __future__ import absolute_import
import logging
import datetime

from tornado import gen
from scrapy.exceptions import NotConfigured

from .utils import tt_coroutine, json_encode


logger = logging.getLogger(__name__)


class MotorPipeline(object):
    """
    This pipeline exports tems to MongoDB using async mongo
    driver (motor). Interaction with MongoDB doesn't block
    the event loop.

    On start it creates object in 'jobs' collection and sets
    spider.motor_job_id to the ID of this job.

    If MOTOR_PIPELINE_JOBID_KEY option is set, job id is added to
    each stored item under the specified key name.
    """
    ITEMS_COLLECTION = 'items'
    JOBS_COLLECTION = 'jobs'

    def __init__(self, crawler):
        try:
            import motor
        except ImportError:
            logger.info("MotorPipeline is disabled because motor Python package "
                        "is not available")
            raise NotConfigured

        self.crawler = crawler
        opts = self.crawler.settings
        if not opts.getbool('MOTOR_PIPELINE_ENABLED', False):
            raise NotConfigured

        self.job_id_key = opts.get('MOTOR_PIPELINE_JOBID_KEY')
        self.db_uri = opts.get('MOTOR_PIPELINE_URI', 'mongodb://localhost:27017')
        db_name = opts.get('MOTOR_PIPELINE_DB_NAME', 'motor_exporter')

        self.client = motor.MotorClient(self.db_uri)
        self.items_table = self.client[db_name][self.ITEMS_COLLECTION]
        self.jobs_table = self.client[db_name][self.JOBS_COLLECTION]
        self.connected = False

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    @tt_coroutine
    def open_spider(self, spider):
        try:
            yield self.items_table.ensure_index(self.job_id_key)

            self.job_id = yield self.jobs_table.insert({
                'started_at': datetime.datetime.utcnow(),
                'spider': spider.name,
            })
            spider.motor_job_id = str(self.job_id)

            logger.info("Crawl job generated id: %s", self.job_id,
                        extra={'crawler': self.crawler})
            self.connected = True
        except Exception:
            self.job_id = None
            logger.error(
                "Can't connect to %s. Items won't be stored.",
                self.db_uri, exc_info=True,
                extra={'crawler': self.crawler},
            )

    @tt_coroutine
    def close_spider(self, spider):
        if self.job_id is None:
            self.client.close()
            return

        # json is to fix an issue with dots in key names
        stats = json_encode(self.crawler.stats.get_stats())

        res = yield self.jobs_table.update(
            {'_id': self.job_id},
            {'$set': {
                'finished_at': datetime.datetime.utcnow(),
                'stats': stats,
            }}
        )
        self.client.close()
        logger.info("Job info %s is saved", self.job_id,
                    extra={'crawler': self.crawler})

    @tt_coroutine
    def process_item(self, item, spider):
        if not self.connected:
            raise gen.Return(item)

        mongo_item = dict(item)
        if self.job_id_key:
            mongo_item[self.job_id_key] = self.job_id

        try:
            yield self.items_table.insert(mongo_item)
            self.crawler.stats.inc_value("motor/items_stored_count")
        except Exception as e:
            self.crawler.stats.inc_value("motor/store_error_count")
            self.crawler.stats.inc_value("motor/store_error_count/" + e.__class__.__name__)
            logger.error("Error storing item", exc_info=True, extra={
                'crawler': self.crawler
            })
        raise gen.Return(item)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import logging
from tornado.ioloop import PeriodicCallback

from arachnado.crawler_process import (
    ArachnadoCrawlerProcess,
    agg_stats_changed,
    CrawlerProcessSignals as CPS,
)
from arachnado.process_stats import ProcessStatsMonitor
from arachnado.wsbase import BaseWSHandler


logger = logging.getLogger(__name__)

#
# class Broker(object):
#     """ Pub/Sub broker """
#     def __init__(self):
#         self.subscribers = defaultdict(set)
#
#     def on(self, event, callback):
#         self.subscribers[event].add(callback)
#
#     def off(self, event, callback):
#         self.subscribers[event].remove(callback)
#
#     def emit(self, event, message):
#         for cb in self.subscribers[event]:
#             cb(message)
#
#
# class WSUpdatesHandler(BaseWSHandler):
#     """ WebSocket handler which sends updates to the client """
#
#     def initialize(self, broker):
#         self.broker = broker
#
#     def on_open(self, *args, **kwargs):
#         pass
#


class Monitor(BaseWSHandler):
    """
    WebSocket handler which pushes CrawlerProcess events to a client.
    """
    engine_signals = [
        CPS.spider_closing, CPS.engine_paused, CPS.engine_resumed,
        CPS.engine_tick, CPS.downloader_enqueued, CPS.downloader_dequeued
    ]

    def initialize(self, crawler_process, opts):
        """
        :param ArachnadoCrawlerProcess crawler_process: crawler process
        """
        self.cp = crawler_process
        self.opts = opts

    def on_open(self):
        logger.debug("new connection")
        self.cp.signals.connect(self.on_stats_changed, agg_stats_changed)
        self.cp.signals.connect(self.on_spider_opened, CPS.spider_opened)
        self.cp.signals.connect(self.on_spider_closed, CPS.spider_closed)

        for signal in self.engine_signals:
            self.cp.signals.connect(self.on_engine_state_changed, signal)

        self.cp.procmon.signals.connect(self.on_process_stats, ProcessStatsMonitor.signal_updated)
        self.write_event("jobs:state", self.cp.jobs)

    def on_close(self):
        logger.debug("connection closed")
        self.cp.signals.disconnect(self.on_stats_changed, agg_stats_changed)
        self.cp.signals.disconnect(self.on_spider_opened, CPS.spider_opened)
        self.cp.signals.disconnect(self.on_spider_closed, CPS.spider_closed)
        for signal in self.engine_signals:
            self.cp.signals.disconnect(self.on_engine_state_changed, signal)
        self.cp.procmon.signals.disconnect(self.on_process_stats, ProcessStatsMonitor.signal_updated)

    def on_spider_opened(self, spider):
        self._send_jobs_state()

    def on_spider_closed(self, spider, reason):
        self._send_jobs_state()

    def on_engine_state_changed(self, crawler):
        self._send_jobs_state()

    def on_tick(self):
        self._send_jobs_state()

    def on_stats_changed(self, changes, crawler):
        # Don't log anything here! Log events are counted by stats collector,
        # so logging a message will trigger more on_stats_changed events.
        crawl_id = crawler.spider.crawl_id
        self.write_event("stats:changed", [crawl_id, changes])

    def on_process_stats(self, stats):
        self.write_event("process:stats", stats)

    def _send_jobs_state(self):
        self.write_event("jobs:state", self.cp.jobs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import urlparse
from scrapy.utils.serialize import ScrapyJSONEncoder

MB = 1024*1024


_encoder = ScrapyJSONEncoder(ensure_ascii=False)
def json_encode(obj):
    """
    Encode a Python object to JSON.
    Unlike standard json.dumps, datetime.datetime objects are supported.
    """
    return _encoder.encode(obj)


def decorate_methods(method_names, decorator):
    """ Class decorator which applies a decorator to all specified methods """
    def _decorator(cls):
        for name in method_names:
            wrapped = decorator(getattr(cls, name))
            setattr(cls, name, wrapped)
        return cls
    return _decorator


def add_scheme_if_missing(url):
    """
    >>> add_scheme_if_missing("example.com/foo")
    'http://example.com/foo'
    >>> add_scheme_if_missing("https://example.com/foo")
    'https://example.com/foo'
    >>> add_scheme_if_missing("//example.com/foo")
    'http://example.com/foo'
    """
    if url.startswith("//"):
        url = "http:" + url
    if "://" not in url:
        url = "http://" + url
    return url


def get_netloc(url):
    """
    >>> get_netloc("example.org/")
    'example.org'
    >>> get_netloc("http://example.org/foo")
    'example.org'
    >>> get_netloc("http://blog.example.org/foo")
    'blog.example.org'
    """
    return urlparse.urlparse(add_scheme_if_missing(url)).netloc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Arachnado Crawling Server.

Usage:
    arachnado [options]
    arachnado show-settings

Options:

  -p --port <port>          Port to use.
  -h --host <host>          Host to bind.
  -c --config <path>        Path to config file.
  -L --loglevel <level>     Set log level.
  --manhole                 Start a manhole server.
  --manhole-port <port>     Manhole server port.
  --manhole-host <host>     Manhole server host.
  --reactor <name>          Set base event loop. Allowed values are
                            "twisted", "tornado" and "auto".
  --debug                   Enable debug mode.
  --version                 Show version information
  -h --help                 Show this help

"""
from __future__ import absolute_import
import os
import sys
import logging

from docopt import docopt
from tornado.ioloop import IOLoop
import tornado.platform.twisted

from arachnado import __version__


logger = logging.getLogger('arachnado')


def setup_event_loop(use_twisted_reactor, debug=True):
    if use_twisted_reactor:
        tornado.platform.twisted.TwistedIOLoop().install()
        if debug:
            print("Using Twisted reactor as a Tornado event loop")
    else:
        tornado.platform.twisted.install()
        IOLoop.instance().set_blocking_log_threshold(0.5)
        if debug:
            print("Using Tornado event loop as a Twisted reactor")


def main(port, host, start_manhole, manhole_port, manhole_host, loglevel, opts):
    from arachnado.handlers import get_application
    from arachnado.crawler_process import ArachnadoCrawlerProcess
    from arachnado import manhole

    settings = {'LOG_LEVEL': loglevel}
    crawler_process = ArachnadoCrawlerProcess(settings)

    app = get_application(crawler_process, opts)
    app.listen(int(port), host)
    logger.info("Arachnado v%s is started on %s:%s" % (__version__, host, port))

    if start_manhole:
        manhole.start(manhole_port, manhole_host, {'cp': crawler_process})
        logger.info("Manhole server is started on %s:%s" % (
            manhole_host, manhole_port))

    crawler_process.start(stop_after_crawl=False)


def _settings(args):
    from arachnado.options import load_settings, ensure_bool

    if args['--config']:
        path = os.path.expanduser(args['--config'])
        assert os.path.exists(path)
        config_files = [path]
    else:
        config_files = []

    def _overrides(section, mapping):
        return [
            [section, ini_key, args[args_key]]
            for ini_key, args_key in mapping.items()
        ]

    overrides = _overrides('arachnado', {
        'port': '--port',
        'host': '--host',
        'reactor': '--reactor',
        'loglevel': '--loglevel',
        'debug': '--debug',
    }) + _overrides('arachnado.manhole', {
        'enabled': '--manhole',
        'port': '--manhole-port',
        'host': '--manhole-host',
    })
    opts = load_settings(config_files, overrides)
    ensure_bool(opts, 'arachnado', 'debug')
    ensure_bool(opts, 'arachnado.storage', 'enabled')
    ensure_bool(opts, 'arachnado.manhole', 'enabled')
    return opts


def run():
    args = docopt(__doc__, version=__version__)
    opts = _settings(args)

    if args['show-settings']:
        from pprint import pprint
        pprint(opts)
        sys.exit(1)

    _reactor = opts['arachnado']['reactor']
    if _reactor not in {'twisted', 'tornado', 'auto'}:
        raise ValueError("Invalid 'reactor': %r" % _reactor)

    if _reactor == 'auto':
        # manhole doesn't work well when run on tornado event loop
        use_twisted_reactor = opts['arachnado.manhole']['enabled']
    else:
        use_twisted_reactor = _reactor == "twisted"

    setup_event_loop(
        use_twisted_reactor=use_twisted_reactor,
        debug=opts['arachnado']['debug']
    )
    main(
        port=int(opts['arachnado']['port']),
        host=opts['arachnado']['host'],
        start_manhole=opts['arachnado.manhole']['enabled'],
        manhole_port=int(opts['arachnado.manhole']['port']),
        manhole_host=opts['arachnado.manhole']['host'],
        loglevel=opts['arachnado']['loglevel'],
        opts=opts,
    )


if __name__ == "__main__":
    run()



import pymongo
from scrapy.http import Headers
from scrapy.responsetypes import responsetypes


class ArachnadoCacheStorage(object):
    """A read-only cache storage that uses Arachnado MongoDB for retrieving
    responses"""

    def __init__(self, settings):
        self.db_name = settings.get('MOTOR_PIPELINE_DB_NAME')
        self.db_uri = settings.get('MOTOR_PIPELINE_URI')

    def open_spider(self, spider):
        self.db = pymongo.MongoClient(self.db_uri)
        self.col = self.db[self.db_name]['items']
        self.col.ensure_index('url')

    def close_spider(self, spider):
        self.db.close()

    def retrieve_response(self, spider, request):
        doc = self.col.find_one({'url': request.url})
        if doc is None:
            return  # not cached
        status = 200
        url = doc['url']
        headers = Headers(doc['headers'])
        body = doc['body'].encode('utf-8')
        respcls = responsetypes.from_args(headers=headers, url=url)
        response = respcls(url=url, headers=headers, status=status, body=body)
        return response

    def store_response(self, spider, request, response):
        pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import json
from tornado import web


class ApiHandler(web.RequestHandler):
    """ Base handler for JSON APIs """

    def prepare(self):
        if self.request.headers.get("Content-Type", "").startswith("application/json"):
            self.json_args = json.loads(self.request.body)
            self.is_json = True
        else:
            self.json_args = None
            self.is_json = False


class NoEtagsMixin(object):
    """ A mixin to fix browser caching of static files referred from a page """
    def compute_etag(self):
        return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import re

from scrapy.utils.misc import walk_modules
from scrapy.utils.spider import iter_spider_classes
from tornado.web import Application, RequestHandler, url, HTTPError

from arachnado.utils import json_encode
from arachnado.spider import create_crawler, CrawlWebsiteSpider
from arachnado.monitor import Monitor
from arachnado.handler_utils import ApiHandler, NoEtagsMixin

at_root = lambda *args: os.path.join(os.path.dirname(__file__), *args)


def get_application(crawler_process, opts):
    context = {
        'crawler_process': crawler_process,
        'opts': opts,
    }
    debug = opts['arachnado']['debug']

    handlers = [
        url(r"/", Index, context, name="index"),
        url(r"/help", Help, context, name="help"),
        url(r"/crawler/start", StartCrawler, context, name="start"),
        url(r"/crawler/stop", StopCrawler, context, name="stop"),
        url(r"/crawler/pause", PauseCrawler, context, name="pause"),
        url(r"/crawler/resume", ResumeCrawler, context, name="resume"),
        url(r"/crawler/status", CrawlerStatus, context, name="status"),
        url(r"/ws-updates", Monitor, context, name="ws"),
    ]
    return Application(
        handlers=handlers,
        template_path=at_root("templates"),
        compiled_template_cache=not debug,
        static_hash_cache=not debug,
        static_path=at_root("static"),
        # no_keep_alive=True,
        compress_response=True,
    )


def get_spider_cls(url, spider_packages,
                   default=CrawlWebsiteSpider):
    """
    Return spider class based on provided url.

    :param url: if it looks like `spider://spidername` it tries to load spider
        named `spidername`, otherwise it returns default spider class
    :param spider_packages: a list of package names that will be searched for
        spider classes
    :param default: the class that is returned when `url` doesn't start with
        `spider://`
    """
    if url.startswith('spider://'):
        spider_name = url[len('spider://'):]
        return find_spider_cls(spider_name, spider_packages)
    return default


def find_spider_cls(spider_name, spider_packages):
    """
    Find spider class which name is equal to `spider_name` argument

    :param spider_name: spider name to look for
    :param spider_packages: a list of package names that will be searched for
        spider classes
    """
    for package_name in spider_packages:
        for module in walk_modules(package_name):
            for spider_cls in iter_spider_classes(module):
                if spider_cls.name == spider_name:
                    return spider_cls


class BaseRequestHandler(RequestHandler):

    def initialize(self, crawler_process, opts):
        """
        :param arachnado.crawler_process.ArachnadoCrawlerProcess
            crawler_process:
        """
        self.crawler_process = crawler_process
        self.opts = opts

    def render(self, *args, **kwargs):
        proc_stats = self.crawler_process.procmon.get_recent()
        kwargs['initial_process_stats_json'] = json_encode(proc_stats)
        return super(BaseRequestHandler, self).render(*args, **kwargs)


class Index(NoEtagsMixin, BaseRequestHandler):

    def get(self):
        jobs = self.crawler_process.jobs
        initial_data_json = json_encode({"jobs": jobs})
        return self.render("index.html", initial_data_json=initial_data_json)


class Help(BaseRequestHandler):
    def get(self):
        return self.render("help.html")


class StartCrawler(ApiHandler, BaseRequestHandler):
    """
    This endpoint starts crawling for a domain.
    """
    def crawl(self, domain):
        storage_opts = self.opts['arachnado.storage']
        settings = {
            'MOTOR_PIPELINE_ENABLED': storage_opts['enabled'],
            'MOTOR_PIPELINE_DB_NAME': storage_opts['db_name'],
            'MOTOR_PIPELINE_DB': storage_opts['db_name'],
            'MOTOR_PIPELINE_URI': storage_opts['uri'],
        }
        spider_cls = get_spider_cls(domain, self._get_spider_package_names())

        if spider_cls is not None:
            self.crawler = create_crawler(settings, spider_cls=spider_cls)
            self.crawler_process.crawl(self.crawler, domain=domain)
            return True
        return False

    def post(self):
        if self.is_json:
            domain = self.json_args['domain']
            if self.crawl(domain):
                self.write({"status": "ok",
                            "job_id": self.crawler.spider.crawl_id})
            else:
                self.write({"status": "error"})
        else:
            domain = self.get_body_argument('domain')
            if self.crawl(domain):
                self.redirect("/")
            else:
                raise HTTPError(400)

    def _get_spider_package_names(self):
        return [name for name in re.split(
            '\s+', self.opts['arachnado.scrapy']['spider_packages']
        ) if name]


class _ControlJobHandler(ApiHandler, BaseRequestHandler):
    def control_job(self, job_id):
        raise NotImplementedError

    def post(self):
        if self.is_json:
            job_id = int(self.json_args['job_id'])
            self.control_job(job_id)
            self.write({"status": "ok"})
        else:
            job_id = int(self.get_body_argument('job_id'))
            self.control_job(job_id)
            self.redirect("/")


class StopCrawler(_ControlJobHandler):
    """ This endpoint stops a running job. """
    def control_job(self, job_id):
        self.crawler_process.stop_job(job_id)


class PauseCrawler(_ControlJobHandler):
    """ This endpoint pauses a job. """
    def control_job(self, job_id):
        self.crawler_process.pause_job(job_id)


class ResumeCrawler(_ControlJobHandler):
    """ This endpoint resumes a paused job. """
    def control_job(self, job_id):
        self.crawler_process.resume_job(job_id)


class CrawlerStatus(BaseRequestHandler):
    """ Status for one or more jobs. """
    def get(self):
        crawl_ids_arg = self.get_argument('crawl_ids', '')

        if crawl_ids_arg == '':
            jobs = self.crawler_process.get_jobs()
        else:
            crawl_ids = set(map(int, crawl_ids_arg.split(',')))
            jobs = [job for job in self.crawler_process.get_jobs()
                    if job['id'] in crawl_ids]

        self.write(json_encode({"jobs": jobs}))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
from setuptools import setup, find_packages

with open("README.rst") as f:
    long_description = f.read()

with open("CHANGES.rst") as f:
    long_description += "\n\n" + f.read()


def get_version():
    fn = os.path.join(os.path.dirname(__file__), "arachnado", "__init__.py")
    with open(fn) as f:
        return re.findall("__version__ = '([\d\.]+)'", f.read())[0]


setup(
    name='arachnado',
    version=get_version(),
    url='https://github.com/TeamHG-Memex/arachnado',
    description='Scrapy-based Web Crawler with an UI',
    long_description=long_description,
    author='Mikhail Korobov',
    author_email='kmike84@gmail.com',
    license='MIT',
    packages=find_packages(),
    package_data={
        'arachnado': [
            "settings/*.conf",
            "templates/*.html",
            "static/css/*.css",
            
            "static/build/*.css",
            "static/build/*.js",

            "static/js/*.jsx",
            "static/js/*.js",
            "static/js/components/*.jsx",
            "static/js/components/*.js",
            "static/js/pages/*.jsx",
            "static/js/pages/*.js",
            "static/js/stores/*.jsx",
            "static/js/stores/*.js",
            "static/js/utils/*.jsx",
            "static/js/utils/*.js",
        ]
    },

    zip_safe=False,
    entry_points={
        'console_scripts': ['arachnado = arachnado.__main__:run']
    },
    classifiers=[
        'Framework :: Scrapy',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=[
        'scrapy >= 1.0.0',
        'Twisted >= 12',
        'psutil >= 2.2',
        'tornado >= 4.2',
        'docopt >= 0.6',
        'service_identity',
    ],
    extras_requires={
        'mongo': ['motor >= 0.4.1']
    }
)



# -*- coding: utf-8 -*-
"""
Signals class.
"""
from __future__ import absolute_import


class Signal(object):
    def __init__(self, name, supports_defer):
        self.name = name
        self.supports_defer = supports_defer

    def __repr__(self):
        return "%s(%r, supports_defer=%r)" % (
            self.__class__.__name__, self.name, self.supports_defer
        )
